"""Workarounds for missing TFLite support for rfft and stft and tf.signal.frame.

based on:
https://github.com/tensorflow/magenta/tree/master/magenta/music
as posted in https://github.com/tensorflow/tensorflow/issues/27303
and https://gist.github.com/padoremu/8288b47ce76e9530eb288d4eec2e0b4d
"""
import math

import tensorflow as tf
import numpy as np


def _rdft_matrix(dft_length):
    """Return a precalculated DFT matrix for positive frequencies only

    Args:
        dft_length (int) - DFT length

    Returns
        rdft_mat (k x n tensor) - precalculated DFT matrix rows are frequencies
            columns are samples, k dimension is dft_length // 2 +1 bins long
    """
    # freq bins
    k = np.arange(0, dft_length // 2 + 1)
    # Samples
    n = np.arange(0, dft_length)
    # complex frequency vector (now normalised to 2 pi)
    omega = -1j * 2.0 * np.pi / dft_length * k
    # complex phase, compute a matrix of value for the complex phase for each sample
    # location (n) and each freq bin (k) outer product If the two vectors have dimensions
    # k and n, then their outer product is an k × n matrix
    phase = np.outer(omega, n)
    # return transposed ready for matrix multiplication
    return np.exp(phase).astype(np.complex64).T


@tf.function
def _rdft(signal, dft_length):
    """DFT for real signals.
    Calculates the onesided dft, assuming real signal implies complex congugaqe symetry,
    hence only onesided DFT is returned.

    Args:
        signal (tensor) signal to transform, assumes that the last dimension is the time dimension
            signal can be framed, e.g. (1, 40, 1024) for a single batch of 40 frames of
            length 1024
        dft_length (int) - DFT length

    Returns:
        spec_real (float32 tensor) - real part of spectrogram, e.g. (1, 40, 513) for a 1024 length dft
        spec_imag (float32 tensor) - imag part of spectrogram, e.g. (1, 40, 513) for a 1024 length dft
    """
    # calculate the positive frequency atoms, and tell tensorflow this is a constant.
    rdft_mat = _rdft_matrix(dft_length)

    # tflite doest support complex types so split into real and imaginary:
    rdft_mat_real = tf.constant(np.real(rdft_mat))
    rdft_mat_imag = tf.constant(np.imag(rdft_mat))

    frame_length = tf.shape(signal)[-1]
    # Right-padding, in case the frame length and DFT lenght are different,
    # pad the signal on the right hand side of the frame
    pad_values = tf.concat(
        [tf.zeros([tf.rank(signal) - 1, 2], tf.int32), [[0, dft_length - frame_length]]], axis=0
    )

    signal_padded = tf.pad(signal, pad_values)

    # matrix multiplying real and imag seperately is faster than using complex types.
    spec_real = tf.matmul(signal_padded, rdft_mat_real)
    spec_imag = tf.matmul(signal_padded, rdft_mat_imag)
    spectrogram = tf.stack([spec_real, spec_imag], axis=-1)

    return spectrogram


def fixed_frame(signal, frame_length, frame_step):
    """tflite-compatible tf.signal.frame for fixed-size input.

    Args:
        signal: Tensor containing signal(s).
        frame_length: Number of samples to put in each frame.
        frame_step: Sample advance between successive frames.

    Returns:
        A new tensor where the last axis (or first, if first_axis) of input
        signal has been replaced by a (num_frames, frame_length) array of individual
        frames where each frame is drawn frame_step samples after the previous one.

    Raises:
        ValueError: if signal has an undefined axis length.  This routine only
        supports framing of signals whose shape is fixed at graph-build time.
    """
    signal_shape = list(signal.shape)
    length_samples = signal_shape[-1]

    if length_samples <= 0:
        raise ValueError("fixed framing requires predefined constant signal length")
    # the number of whole frames
    num_frames = max(0, 1 + (length_samples - frame_length) // frame_step)

    # define the output_shape, if we recieve a None dimension, replace with 1
    outer_dimensions = [dim if dim else 1 for dim in signal_shape[:-1]]
    # outer_dimensions = signal_shape[:-1]
    output_shape = outer_dimensions + [num_frames, frame_length]

    # Currently tflite's gather only supports axis==0, but that may still
    # work if we want the last of 1 axes.
    gather_axis = len(outer_dimensions)

    # subframe length is the largest int that as a common divisor of the frame
    # length and hop length. We will slice the signal up into these subframes
    # in order to then construct the frames.
    subframe_length = math.gcd(frame_length, frame_step)
    subframes_per_frame = frame_length // subframe_length
    subframes_per_hop = frame_step // subframe_length
    num_subframes = length_samples // subframe_length

    # define the subframe shape and the trimmed audio length, removeing any unused
    # excess audio, so subframe fit exactly.
    subframe_shape = outer_dimensions + [num_subframes, subframe_length]
    trimmed_input_size = outer_dimensions + [num_subframes * subframe_length]
    # slice up the audio into subframes
    subframes = tf.reshape(
        tf.slice(signal, begin=np.zeros(len(signal_shape), np.int32), size=trimmed_input_size),
        subframe_shape,
    )

    # frame_selector is a [num_frames, subframes_per_frame] tensor
    # that indexes into the appropriate frame in subframes. For example:
    # [[0, 0, 0, 0], [2, 2, 2, 2], [4, 4, 4, 4]]
    frame_selector = np.reshape(np.arange(num_frames) * subframes_per_hop, [num_frames, 1])

    # subframe_selector is a [num_frames, subframes_per_frame] tensor
    # that indexes into the appropriate subframe within a frame. For example:
    # [[0, 1, 2, 3], [0, 1, 2, 3], [0, 1, 2, 3]]
    subframe_selector = np.reshape(np.arange(subframes_per_frame), [1, subframes_per_frame])

    # Adding the 2 selector tensors together produces a [num_frames,
    # subframes_per_frame] tensor of indices to use with tf.gather to select
    # subframes from subframes. We then reshape the inner-most subframes_per_frame
    # dimension to stitch the subframes together into frames. For example:
    # [[0, 1, 2, 3], [2, 3, 4, 5], [4, 5, 6, 7]].
    selector = frame_selector + subframe_selector
    frames = tf.reshape(
        tf.gather(subframes, selector.astype(np.int32), axis=gather_axis), output_shape
    )

    return frames


def stft_tflite(signal, frame_length, frame_step, fft_length, window_fn, pad_end):
    """tflite-compatible implementation of tf.signal.stft.
    Compute the short-time Fourier transform of a 1D input while avoiding tf ops
    that are not currently supported in tflite (Rfft, Range, SplitV).
    fft_length must be fixed. A Hann window is of frame_length is always
    applied.
    Since fixed (precomputed) framing must be used, signal.shape[-1] must be a
    specific value (so "?"/None is not supported).

    Args:
        signal: 1D tensor containing the time-domain waveform to be transformed.
        frame_length: int, the number of points in each Fourier frame.
        frame_step: int, the number of samples to advance between successive frames.
        fft_length: int, the size of the Fourier transform to apply.
        window_fn: tf.signal.window, the return of backend.get_window_fn(window_name)
        pad_end: bool, if true pads the end with zeros so that signal contains
            an integer number of frames

    Returns:
        spectrogram: Two (num_frames, fft_length) tensors containing the real and
            imaginary parts of the short-time Fourier transform of the input signal.
    """
    signal = tf.cast(signal, tf.float32)
    if pad_end:
        # the number of whole frames
        # (NOTE: kenders2000), padding is pre-calculated and thus fixed in graph
        length_samples = signal.shape[-1]
        num_steps_round_up = int(np.ceil(length_samples / frame_step))
        pad_amount = (num_steps_round_up * frame_step + frame_length - frame_step) - length_samples
        signal = tf.pad(signal, tf.constant([[0, 0], [0, 0], [0, pad_amount]]))
    # Make the window be shape (1, frame_length) instead of just frame_length
    # in an effort to help the tflite broadcast logic.
    window = tf.reshape(window_fn(frame_length), [1, frame_length])

    framed_signal = fixed_frame(signal, frame_length, frame_step)
    framed_signal *= window

    spectrogram = _rdft(framed_signal, fft_length)

    return spectrogram


@tf.function
def continued_fraction_arctan(x, n=100, dtype=tf.float32):
    """Continued fraction Approximation to the arctan function

        Approximate solution to arctan(x), atan is not a natively supported tflite
        op (or a flex op). n is the number of iterations, the high the more accurate.
        Accuracy is poor when the argument is large.
        https://functions.wolfram.com/ElementaryFunctions/ArcTan/10/

    Args:
        x (tensor) - argument tensor to caclulate arctan of
        n (int) - The number of iterations, large means arctan is more accurate
        dtype (tf.dtype) - tf.float32, or tf.float64

    Returns
        arctan(x) (tensor) - approx value of arctan(x)
    """
    x = tf.cast(x, dtype)
    x2 = x * x
    d = tf.zeros(tf.shape(x), dtype) + tf.cast(n * 2 + 1, dtype)
    for k in tf.range(n, 0.0, -1.0, dtype):
        f = k * 2.0 - 1.0
        d = f + k * k * x2 / d
    return x / d


def atan2_tflite(y, x, n=100, dtype=tf.float32):
    """Approximation to the atan2 function

        atan is not a tflite supported op or flex op, thus this uses an Approximation
        Poor accuracy when either x is very small or y is very large.
        https://en.wikipedia.org/wiki/Atan2

    Args:
        y (tensor) - vertical component of tangent (or imaginary part of number for phase)
        x (tensor) - horizontal component of tanget (or real part of number for phase)
        n (int) - The number of iterations to use for atan approximations,
            larger means arctan is more accurate
        dtype (tf.dtype) - tf.float32, or tf.float64

    Returns
        atan2(x) (tensor) - approx value of atan2(x)
    """
    pi = tf.zeros(tf.shape(x), dtype) + tf.cast(np.pi, dtype)
    zeros = tf.zeros(tf.shape(x), dtype)
    atan2 = continued_fraction_arctan(y / x, n, dtype)
    atan2 = tf.where(x > 0, atan2, atan2)  # implicit
    atan2 = tf.where(tf.logical_and(x < 0.0, y >= 0.0), atan2 + pi, atan2)
    atan2 = tf.where(tf.logical_and(x < 0.0, y < 0.0), atan2 - pi, atan2)
    atan2 = tf.where(tf.logical_and(tf.equal(x, 0.0), y > 0.0), pi, atan2)
    atan2 = tf.where(tf.logical_and(tf.equal(x, 0.0), y < 0.0), -pi, atan2)
    # undefined (return 0)
    atan2 = tf.where(tf.logical_and(tf.equal(x, 0.0), tf.equal(y, 0.0)), zeros, atan2)
    return atan2



"""Augmentation layers.

This module includes augmentation layers that can be applied to audio data and representations.

"""
import tensorflow as tf
from tensorflow.keras.layers import Layer
from tensorflow.keras import backend as K
from . import backend
from .backend import _CH_FIRST_STR, _CH_LAST_STR, _CH_DEFAULT_STR
import numpy as np

__all__ = ['SpecAugment', 'ChannelSwap']


class ChannelSwap(Layer):
    """
    Randomly swap the channel

    Args:
        data_format (`str`): specifies the data format of batch input/output
        **kwargs: Keyword args for the parent keras layer (e.g., `name`)

    Example:
        ::

            input_shape = (2048, 2)  # stereo signal
            n_fft = 1024
            n_hop = n_fft // 2
            kwargs = {
                'sample_rate': 22050,
                'n_freq': n_fft // 2 + 1,
                'n_mels': 128,
                'f_min': 0.0,
                'f_max': 8000,
            }
            model = Sequential()
            model.add(kapre.STFT(n_fft=n_fft, hop_length=n_hop, input_shape=input_shape))
            model.add(Magnitude())
            # (batch, n_frame=3, n_freq=n_fft // 2 + 1, ch=1) and dtype is float
            model.add(ChannelSwap())
            # same shape but with randomly swapped channels.

        ::

            input_shape = (2048, 4)  # mono signal
            model = Sequential()
            model.add(ChannelSwap(input_shape=input_shape))
            # still (2048, 4) but with randomly swapped channels

    """

    def __init__(
        self,
        data_format='default',
        **kwargs,
    ):
        backend.validate_data_format_str(data_format)

        if data_format == _CH_DEFAULT_STR:
            self.data_format = K.image_data_format()
        else:
            self.data_format = data_format

        super(ChannelSwap, self).__init__(**kwargs)

    def call(self, x, training=None):
        """
        Apply random channel-swap augmentation to `x`.

        Args:
            x (`Tensor`): A batch tensor of 1D (signals) or 2D (spectrograms) data
        """
        if training in (None, False):
            return x

        # figure out input data format
        if K.ndim(x) not in (3, 4):
            raise ValueError(
                'ndim of input tensor x should be 3 (batch signal) or 4 (batch spectrogram),'
                'but it is %d' % K.ndim(x)
            )

        if self.data_format == _CH_LAST_STR:
            ch_axis = 3 if K.ndim(x) == 4 else 2
        else:
            ch_axis = 1

        # get swap indices
        n_ch = K.int_shape(x)[ch_axis]
        if n_ch == 1:
            return x
        swap_indices = np.random.permutation(n_ch).tolist()

        # swap and return
        return tf.gather(x, indices=swap_indices, axis=ch_axis)

    def get_config(self):
        config = super(ChannelSwap, self).get_config()
        config.update(
            {
                'data_format': self.data_format,
            }
        )
        return config


class SpecAugment(Layer):
    """
    Apply SpecAugment to a Spectrogram. For more info, check the original paper at:
    https://arxiv.org/abs/1904.08779

    Args:
        freq_mask_param (`int`): Frequency Mask Parameter (F in the paper)
        time_mask_param (ìnt`): Time Mask Parameter (T in the paper)
        n_freq_masks (`int`): Number of frequency masks to apply (mF in the paper). By default is 1.
        n_time_masks (`int`): Number of time masks to apply (mT in the paper). By default is 1.
        mask_value (`float`): Value of the applied masks. By default is 0.
        data_format (`str`): specifies the data format of batch input/output
        **kwargs: Keyword args for the parent keras layer (e.g., `name`)

    Example:
        ::

            input_shape = (2048, 1)  # mono signal

            # We compute the Mel Spectrogram of the input signal
            melgram = kapre.composed.get_melspectrogram_layer(input_shape=input_shape,
                                                  n_fft=1024,
                                                  return_decibel=True,
                                                  n_mels=256,
                                                  input_data_format='channels_last',
                                                  output_data_format='channels_last')


            # Now we define the SpecAugment layer. It will apply 5 masks in the frequency axis,
            # 3 masks in the time axis. The frequency mask param is 5 and the time mask param
            # is 10.
            spec_augment = SpecAugment(freq_mask_param=5,
                                       time_mask_param=10,
                                       n_freq_masks=5,
                                       n_time_masks=3)

            model = Sequential()
            model.add(melgram)

            # Add the spec_augment layer for augmentation
            model.add(spec_augment)
        ::
    """

    def __init__(
        self,
        freq_mask_param,
        time_mask_param,
        n_freq_masks=1,
        n_time_masks=1,
        mask_value=0.0,
        data_format='default',
        **kwargs,
    ):

        backend.validate_data_format_str(data_format)

        super(SpecAugment, self).__init__(**kwargs)

        self.freq_mask_param = freq_mask_param
        self.time_mask_param = time_mask_param
        self.n_freq_masks = n_freq_masks
        self.n_time_masks = n_time_masks
        self.mask_value = mask_value

        if not self.freq_mask_param or not self.time_mask_param:
            raise RuntimeError(
                "Both freq_mask_param and time_mask_param must be defined and different "
                "than zero"
            )

        self.data_format = K.image_data_format() if data_format == _CH_DEFAULT_STR else data_format

    @staticmethod
    def _generate_axis_mask(inputs):
        """
        Generate a mask for the axis provided
        Args:
            inputs (`tuple`): A 3-tuple with the following structure:
                inputs[0] (float `Tensor`): A spectrogram. Its shape is (time, freq, ch) or (ch, time, freq)
                    depending on data_format
                inputs[1] (int): The axis limit. If mask will be applied to time axis it will be `time`, if it will
                    be applied to frequency axis, then it will be `freq`
                inputs[2] (int `Tensor`): The axis indices. We need this Tensor of indices to indicate where to apply
                    the mask.
                inputs[3] (int): The mask param as defined in the original paper, which is the max width of the mask
                    applied.
        Returns:
            (bool `Tensor`): A boolean tensor representing the mask. Its shape is (time, freq, ch) or (ch, time, freq)
                depending on inputs[0] shape (that is, the input spectrogram).
        """
        x, axis_limit, axis_indices, mask_param = inputs

        mask_width = tf.random.uniform(shape=(), maxval=mask_param, dtype=tf.int32)
        mask_start = tf.random.uniform(shape=(), maxval=axis_limit - mask_width, dtype=tf.int32)

        return tf.logical_and(axis_indices >= mask_start, axis_indices <= mask_start + mask_width)

    def _apply_masks_to_axis(self, x, axis, mask_param, n_masks):
        """
        Applies a number of masks (defined by the parameter n_masks) to the spectrogram
        by the axis provided.
        Args:
            x (float `Tensor`): A spectrogram. Its shape is (time, freq, ch) or (ch, time, freq)
                    depending on data_format.
            axis (int): The axis where the masks will be applied
            mask_param (int): The mask param as defined in the original paper, which is the max width of the mask
                    applied to the specified axis.
            n_masks (int): The number of masks to be applied

        Returns:
            (float `Tensor`): The masked spectrogram. Its shape is (time, freq, ch) or (ch, time, freq)
                depending on x shape (that is, the input spectrogram).
        """
        axis_limit = K.int_shape(x)[axis]
        axis_indices = tf.range(axis_limit)

        if axis == 0:
            axis_indices = tf.reshape(axis_indices, (-1, 1, 1))
        elif axis == 1:
            axis_indices = tf.reshape(axis_indices, (1, -1, 1))
        elif axis == 2:
            axis_indices = tf.reshape(axis_indices, (1, 1, -1))
        else:
            raise NotImplementedError(f"Axis parameter must be one of the following: 0, 1, 2")

        # Check if mask_width is greater than axis_limit
        if axis_limit < mask_param:
            raise ValueError(
                "Time and freq axis shapes must be greater than time_mask_param "
                "and freq_mask_param respectively"
            )

        x_repeated = tf.repeat(tf.expand_dims(x, 0), n_masks, axis=0)
        axis_limit_repeated = tf.repeat(axis_limit, n_masks, axis=0)
        axis_indices_repeated = tf.repeat(tf.expand_dims(axis_indices, 0), n_masks, axis=0)
        mask_param_repeated = tf.repeat(mask_param, n_masks, axis=0)

        masks = tf.map_fn(
            elems=(x_repeated, axis_limit_repeated, axis_indices_repeated, mask_param_repeated),
            fn=self._generate_axis_mask,
            dtype=(tf.float32, tf.int32, tf.int32, tf.int32),
            fn_output_signature=tf.bool,
        )

        mask = tf.math.reduce_any(masks, 0)
        return tf.where(mask, self.mask_value, x)

    def _apply_spec_augment(self, x):
        """
        Main method that applies SpecAugment technique by both frequency and
        time axis.
        Args:
            x (float `Tensor`) : A spectrogram. Its shape is (time, freq, ch) or (ch, time, freq)
                    depending on data_format.
        Returns:
            (float `Tensor`): The spectrogram masked by time and frequency axis. Its shape is (time, freq, ch)
                or (ch, time, freq) depending on x shape (that is, the input spectrogram).
        """
        if self.data_format == _CH_LAST_STR:
            time_axis, freq_axis = 0, 1
        else:
            time_axis, freq_axis = 1, 2

        if self.n_time_masks >= 1:
            x = self._apply_masks_to_axis(
                x, axis=time_axis, mask_param=self.time_mask_param, n_masks=self.n_time_masks
            )
        if self.n_freq_masks >= 1:
            x = self._apply_masks_to_axis(
                x, axis=freq_axis, mask_param=self.freq_mask_param, n_masks=self.n_freq_masks
            )
        return x

    def call(self, x, training=None, **kwargs):
        if training in (None, False):
            return x

        if K.ndim(x) != 4:
            raise ValueError(
                'ndim of input tensor x should be 4 (batch spectrogram),' 'but it is %d' % K.ndim(x)
            )

        ch_axis = 1 if self.data_format == 'channels_first' else 3

        if K.int_shape(x)[ch_axis] != 1:
            raise RuntimeError(
                'SpecAugment does not support spectrograms with depth greater than 1'
            )

        return tf.map_fn(
            elems=x, fn=self._apply_spec_augment, dtype=tf.float32, fn_output_signature=tf.float32
        )

    def get_config(self):
        config = super(SpecAugment, self).get_config()
        config.update(
            {
                'freq_mask_param': self.freq_mask_param,
                'time_mask_param': self.time_mask_param,
                'n_freq_masks': self.n_freq_masks,
                'n_time_masks': self.n_time_masks,
                'mask_value': self.mask_value,
                'data_format': self.data_format,
            }
        )
        return config



"""Tflite compatible versions of Kapre layers.

`STFTTflite` is a tflite compatible version of `STFT`. Tflite does not support complex
types, thus real and imaginary parts are returned as an extra (last) dimension.
Ouput shape is now: `(batch, channel, time, re/im)` or `(batch, time, channel, re/im)`.

Because of the change of dimension, Tflite compatible layers are provided to
process the resulting STFT; `MagnitudeTflite` and `PhaseTflite` are layers that
calculate the magnitude and phase respectively from the output of `STFTTflite`.
"""
import tensorflow as tf
from .backend import _CH_FIRST_STR, _CH_LAST_STR, _CH_DEFAULT_STR
from .tflite_compatible_stft import stft_tflite, atan2_tflite

# import non-tflite compatible layers to inheret from.
from .time_frequency import STFT, InverseSTFT, Magnitude, Phase


__all__ = [
    'STFTTflite',
    # 'InverseSTFTTflite',  # NOTE (PK): todo
    'MagnitudeTflite',
    'PhaseTflite',
]


class STFTTflite(STFT):
    """
    A Short-time Fourier transform layer (tflite compatible).

    Ues `stft_tflite` from tflite_compatible_stft.py, this contains a tflite
    compatible stft (using a rdft), and `fixed_frame()` to window the audio.
    Tflite does not cope with comple types so real and imaginary parts are stored in extra dim.
    Ouput shape is now: (batch, channel, time, re/im) or (batch, time, channel, re/im).
    `MagnitudeTflite`, and `PhaseTflite` are versions of the `Magnitude` and `Phase`
    layers that account for this extra dimensionality. Currently this layer is
    restricted to a batch size of one, for training use the `STFT` layer, and
    once complete transfer the weights to a new model, replacing the `STFT` layer
    with the `STFTTflite` layer and `Magnitude` and `Phase` layers with
    `MagnitudeTflite` and `PhaseTflite` layers.

    Additionally, it reshapes the output to be a proper 2D batch.

    If `output_data_format == 'channels_last'`, the output shape is `(batch, time, freq, channel, re/imag)`
    If `output_data_format == 'channels_first'`, the output shape is `(batch, channel, time, freq, re/imag)`

    Args:
        n_fft (int): Number of FFTs. Defaults to `2048`
        win_length (int or None): Window length in sample. Defaults to `n_fft`.
        hop_length (int or None): Hop length in sample between analysis windows. Defaults to `n_fft // 4` following Librosa.
        window_name (str or None): *Name* of `tf.signal` function that returns a 1D tensor window that is used in analysis.
            Defaults to `hann_window` which uses `tf.signal.hann_window`.
            Window availability depends on Tensorflow version. More details are at `kapre.backend.get_window()`.
        pad_begin (bool): Whether to pad with zeros along time axis (length: win_length - hop_length). Defaults to `False`.
        pad_end (bool): Whether to pad with zeros at the finishing end of the signal.
        input_data_format (str): the audio data format of input waveform batch.
            `'channels_last'` if it's `(batch, time, channels)` and
            `'channels_first'` if it's `(batch, channels, time)`.
            Defaults to the setting of your Keras configuration. (`tf.keras.backend.image_data_format()`)
        output_data_format (str): The data format of output STFT.
            `'channels_last'` if you want `(batch, time, frequency, channels)` and
            `'channels_first'` if you want `(batch, channels, time, frequency)`
            Defaults to the setting of your Keras configuration. (`tf.keras.backend.image_data_format()`)

        **kwargs: Keyword args for the parent keras layer (e.g., `name`)

    Example:
        ::

            input_shape = (2048, 1)  # mono signal
            model = Sequential()  # tflite compatible model
            model.add(kapre.STFTTflite(n_fft=1024, hop_length=512, input_shape=input_shape))
            # now the shape is (batch, n_frame=3, n_freq=513, ch=1, re/im=2)
            # and the dtype is real

    """

    def call(self, x):
        """
        Compute STFT of the input signal. If the `time` axis is not the last axis of `x`, it should be transposed first.

        Args:
            x (float `Tensor`): batch of audio signals, (batch, ch, time) or (batch, time, ch) based on input_data_format

        Return:
            (real `Tensor`): A STFT representation of x in a 2D batch shape. The last dimension is size two and contains
            the real and imaginary parts of the stft.
            Its shape is (batch, time, freq, ch, 2) or (batch. ch, time, freq, 2) depending on `output_data_format` and
            `time` is the number of frames, which is `((len_src + (win_length - hop_length) / hop_length) // win_length )`
            if `pad_end` is `True`. `freq` is the number of fft unique bins, which is `n_fft // 2 + 1` (the unique components of the FFT).
        """
        waveforms = x  # (batch, ch, time) if input_data_format == 'channels_first'.
        # (batch, time, ch) if input_data_format == 'channels_last'.

        # this is needed because tf.signal.stft lives in channels_first land.
        if self.input_data_format == _CH_LAST_STR:
            waveforms = tf.transpose(
                waveforms, perm=(0, 2, 1)
            )  # always (batch, ch, time) from here

        if self.pad_begin:
            waveforms = tf.pad(
                waveforms, tf.constant([[0, 0], [0, 0], [int(self.n_fft - self.hop_length), 0]])
            )
        stfts = stft_tflite(
            waveforms,
            frame_length=self.win_length,
            frame_step=self.hop_length,
            fft_length=self.n_fft,
            window_fn=self.window_fn,
            pad_end=self.pad_end,
        )  # (batch, ch, time, freq, re/imag)

        if self.output_data_format == _CH_LAST_STR:
            # tflite compatible stft produces real and imag in 1st dim
            stfts = tf.transpose(stfts, perm=(0, 2, 3, 1, 4))  # (batch, t, f, ch, re/im)

        return stfts


class MagnitudeTflite(Magnitude):
    """Compute the magnitude of the input (tflite compatible).

    The input is a real tensor, the last dimension has a size of `2`
    representing real and imaginary parts respectively.

    Example:
        ::

            input_shape = (2048, 1)  # mono signal
            model = Sequential()
            model.add(kapre.STFTTflite(n_fft=1024, hop_length=512, input_shape=input_shape))
            mode.add(MagnitudeTflite())
            # now the shape is (batch, n_frame=3, n_freq=513, ch=1) and dtype is float

    """

    def call(self, x):
        """
        Args:
            x (real or complex `Tensor`): input is real tensor whose last
                dimension has a size of `2` and represents real and imaginary
                parts

        Returns:
            (float `Tensor`): magnitude of `x`
        """
        return tf.norm(x, ord='euclidean', axis=-1)


class PhaseTflite(Phase):
    """Compute the phase of the complex input in radian, resulting in a float tensor (tflite compatible).

    Note TF lite does not natively support atan, used in tf.math.angle, so an
    approximation is provided. You may want to use this approximation if you
    generate data using a non-tf-lite compatible STFT (faster) but want the same
    approximations in the training data.

    Args:
        approx_atan_accuracy (`int`): if `None` will use `tf.math.angle()` to
            calculate the phase accurately. If an `int` this is the number of
            iterations to calculate the approximate `atan()` using a tflite compatible
            method. the higher the number the more accurate e.g.
            `approx_atan_accuracy=29000`. You may want to experiment with adjusting
            this number: trading off accuracy with inference speed.

    Example:
        ::

            input_shape = (2048, 1)  # mono signal
            model = Sequential()
            model.add(kapre.STFTTflite(n_fft=1024, hop_length=512, input_shape=input_shape))
            model.add(PhaseTflite(approx_atan_accuracy=5000))
            # now the shape is (batch, n_frame=3, n_freq=513, ch=1) and dtype is float

    """

    def call(self, x):
        """
        Args:
            x (real): input is real tensor with five
                dimensions (last dim is re/imag)

        Returns:
            (float `Tensor`): phase of `x` (Radian)
        """
        return atan2_tflite(x[:, :, :, :, 1], x[:, :, :, :, 0], n=self.approx_atan_accuracy)
